from typing import Annotated, Any, Literal

from polars import DataFrame
from pydantic import BaseModel, ConfigDict, Field

from ..interface import SourceConfig, SourceConfigArguments, SourceConnectionArguments


class FakerConnectionArguments(SourceConnectionArguments):
    faker: list[dict[str, Any]]


class FakerConfigArguments(SourceConfigArguments):

    type: Annotated[Literal["Faker"], Field(default="Faker")] = "Faker"
    connection: FakerConnectionArguments


class FakerConfig(SourceConfig):
    """The connection configuration class used for faker sources"""

    data: DataFrame

    def __init__(self, arguments: FakerConfigArguments):
        super().__init__(arguments)
        self.data = DataFrame(arguments.connection.faker or [])



from typing import Literal

import polars as pl

PolarTypeLiterals = Literal[
    "Decimal",
    "Float32",
    "Float64",
    "Int8",
    "Int16",
    "Int32",
    "Int64",
    "UInt8",
    "UInt16",
    "UInt32",
    "UInt64",
    "Date",
    "Datetime",
    "Duration",
    "Time",
    "List",
    "Array",
    "Categorical",
    "Enum",
    "Struct",
    "Object",
    "Utf8",
    "String",
]


def get_polar_type_from_literal(type: str) -> pl.DataType:
    return getattr(pl, type)


PolarTypeOverridesParameters = dict[str, PolarTypeLiterals]


class PolarTypeOverrides(dict[str, pl.DataType]):
    def __init__(self, overrides: PolarTypeOverridesParameters):
        super().__init__(overrides)
        for key, value in overrides.items():
            self[key] = get_polar_type_from_literal(value)



from typing import Annotated, Literal, Optional

from pydantic import BaseModel, Field

from ...config.polar_types import PolarTypeOverrides, PolarTypeOverridesParameters
from ..interface import SourceSync, SourceSyncArguments

DATAMODELS_WITHOUT_SOFT_DELETE = (
    "NutanixCluster",
    "NutanixClusterHost",
    "NutanixVM",
    "NutanixVMDisk",
    "NutanixNetwork",
    "NutanixNetworkInterface",
)


class SoftDeleteArguments(BaseModel):
    enabled: Annotated[bool, Field(default=False)]
    field: Annotated[str, Field(min_length=1, default="status")]
    active_value: Annotated[str, Field(min_length=1, default="enabled")]
    inactive_value: Annotated[str, Field(min_length=1, default="inactive")]


class ItopSyncArguments(SourceSyncArguments):
    datamodel: Annotated[str, Field(min_length=1)]
    oql_key: Annotated[str, Field(min_length=1)]
    soft_delete: Annotated[Optional[SoftDeleteArguments], Field(default=None)]
    link_columns: Annotated[list[str], Field(default=[])]
    comparison_columns: Annotated[list[str], Field(min_length=1)]
    unique_columns: Annotated[list[str], Field(min_length=1)]
    skip_columns: Annotated[Optional[list[str]], Field(default=[])]
    foreign_key_columns: Annotated[Optional[list[str]], Field(default=[])]
    type_overrides: Annotated[Optional[PolarTypeOverridesParameters], Field(default={})]

    type: Annotated[Literal["Itop"], Field(default="Itop")] = "Itop"


class ItopSync(SourceSync):
    """The configuration class used for iTop sources"""

    datamodel: str
    oql_key: str
    soft_delete: Optional[SoftDeleteArguments]
    link_columns: list[str]
    comparison_columns: list[str]
    unique_columns: list[str]
    skip_columns: list[str]
    foreign_key_columns: list[str]
    type_overrides: PolarTypeOverrides

    def __init__(self, arguments: ItopSyncArguments):
        super().__init__(arguments)

        self.datamodel = arguments.datamodel
        self.oql_key = arguments.oql_key
        self.comparison_columns = arguments.comparison_columns
        self.unique_columns = arguments.unique_columns
        self.skip_columns = arguments.skip_columns
        self.soft_delete = arguments.soft_delete
        self.link_columns = arguments.link_columns
        self.foreign_key_columns = arguments.foreign_key_columns
        self.type_overrides = PolarTypeOverrides(arguments.type_overrides)

        # TODO: This can be made redundant by adding the model to the soft delete and only allowing the values to be models that support soft delete
        if arguments.soft_delete is not None:
            soft_delete_not_supported = (
                arguments.soft_delete.enabled
                and arguments.datamodel in DATAMODELS_WITHOUT_SOFT_DELETE
            )
            if soft_delete_not_supported:
                raise Exception(
                    f"Soft delete is not supported for {arguments.datamodel} datamodel"
                )



from typing import Annotated, Any, Literal, Optional, Union

from pydantic import BaseModel, ConfigDict, Field, ValidationError, model_validator
from pydantic_core import ErrorDetails

from ...sources import (
    CsvConfigArguments,
    FakerConfigArguments,
    ItopConfigArguments,
    ItopSyncArguments,
    MongodbConfigArguments,
    MongodbSyncArguments,
    MysqlConfigArguments,
    MysqlSyncArguments,
    PostgresConfigArguments,
    PostgresSyncArguments,
    RestConfigArguments,
    RestSyncArguments,
    Sources,
    SqlserverConfigArguments,
    SqlserverSyncArguments,
    StaticConfigArguments,
    XmlConfigArguments,
    XmlSyncArguments,
)
from ...transformers import TransformerSchemas

SourceConfigArguments = list[
    Union[
        (
            StaticConfigArguments,
            ItopConfigArguments,
            MongodbConfigArguments,
            CsvConfigArguments,
            FakerConfigArguments,
            MysqlConfigArguments,
            PostgresConfigArguments,
            RestConfigArguments,
            SqlserverConfigArguments,
            XmlConfigArguments,
        )
    ]
]

SourceSyncArguments = Union[
    (
        ItopSyncArguments,
        MongodbSyncArguments,
        MysqlSyncArguments,
        PostgresSyncArguments,
        RestSyncArguments,
        SqlserverSyncArguments,
        XmlSyncArguments,
        # Necessary for sources that don't have a sync configuration
        dict,
    )
]


class ComparisonColumnV1(BaseModel):
    model_config = ConfigDict(extra="forbid")
    name: Annotated[str, Field(min_length=1)]
    type: Annotated[str, Field(min_length=1)]
    unique: Annotated[bool, Field(default=False)]


ComparisonColumns = Union[list[ComparisonColumnV1], dict[str, str]]
OptionalTransformers = Annotated[
    list[Annotated[TransformerSchemas, Field(discriminator="transformer")]],
    Field(default=[]),
]


class V1Sync(BaseModel):
    source_to_type: Annotated[
        dict[str, str],
        Field(
            default={},
            description="Should never be provided by the user. Is automatically populated at validation time.",
        ),
    ]
    location: Annotated[
        tuple[str, ...],
        Field(
            default=(),
            description="Should never be provided by the user. Is automatically populated at validation time.",
        ),
    ]

    source: Annotated[str, Field(min_length=1)]
    destination: Annotated[str, Field(min_length=1)]
    sourceConfig: Annotated[SourceSyncArguments, Field()]
    destinationConfig: Annotated[SourceSyncArguments, Field()]
    comparisonColumns: Annotated[
        ComparisonColumns,
        Field(min_length=1),
    ]
    sourceTransformers: OptionalTransformers
    destinationTransformers: OptionalTransformers
    insertionTransformers: OptionalTransformers
    deletionTransformers: OptionalTransformers

    @model_validator(mode="before")
    def validate_sources(cls, values):
        """Makes sure that each source configuration is only validated against one source type, the one that matches the 'type' field.
        This is necessary since without it the static jsonschema validation will display errors from ALL source sync config types if the the config isn't fulfilling any of them, which can be confusing for the user.
        """
        source_to_type = values.get("source_to_type", {})
        if not source_to_type:
            raise ValueError(
                "source_to_type is missing, something is wrong with the validation setup."
            )

        errors = []
        for direction in ["source", "destination"]:
            errors.extend(cls.validate_sync_config(direction, values))

        if errors:
            raise ValidationError.from_exception_data(cls.__name__, errors)
        return values

    @classmethod
    def validate_sync_config(
        cls, direction: Literal["source", "destination"], values: dict
    ):
        """Validates the source or destination sync configuration against the correct source. Makes sure that the configuration matches exactly one source type."""
        errors = []
        source_name = values.get(direction, None)
        source_type = values.get("source_to_type", {}).get(source_name, None)
        source_sync_config = values.get(f"{direction}Config", None)
        if source_name and source_type and source_sync_config:
            registrated_source = Sources.sources.get(source_type, None)
            location = (*values["location"], f"{direction}Config")
            try:
                registrated_source.cls.SyncArgumentsClass(
                    direction=direction,
                    name=source_name,
                    location=location,
                    type=source_type,
                    **source_sync_config,
                )
            except ValidationError as e:
                errors.extend(adjust_validation_error_location(e, location))

        return errors


class BeetlConfigSchemaV1(BaseModel):
    """Represents the configuration as supplied by the user. This class is used to validate the configuration against the static jsonschema and the dynamic beetl validation rules."""

    version: Literal["V1"]
    sources: Annotated[SourceConfigArguments, Field(min_items=1)]
    sync: Annotated[list[V1Sync], Field(min_items=1)]

    @model_validator(mode="before")
    def populate_validation_values_in_nested_types(cls, values):
        """Makes sure that a map of source names to source types is available in each sync dict. This is neccessary for the validation of the source and destination config to identify what class each config should be validated against. These values are automatically generated and should never be provided by the user."""
        if not values.get("sync", []) or not values.get("sources", []):
            return values
        source_to_type_dictionary = {}
        for source in values["sources"]:
            source_to_type_dictionary[source["name"]] = source["type"]
        for index, sync in enumerate(values["sync"]):
            sync["source_to_type"] = source_to_type_dictionary
            sync["location"] = ("sync", str(index))
        return values

    @model_validator(mode="before")
    def validate_sources(cls, values):
        """Makes sure that each source in the sources list is valid against the correct source config class."""
        sources: list[dict[str, Any]] = values.get("sources", [])
        errors = []
        for i, source in enumerate(sources):
            source_type = source.get("type", None)
            if source_type is None:
                raise ValueError(
                    f"sources.{i}.type is missing, valid sources are {list(Sources.sources.keys())}"
                )

            registrated_source = Sources.sources.get(source_type, None)
            if registrated_source is None:
                raise ValueError(
                    f'sources.{i}.type "{source_type}" is not recognized, valid sources are {list(Sources.sources.keys())}'
                )
            location = ("sources", str(i))
            try:
                registrated_source.cls.ConfigArgumentsClass(location=location, **source)
            except ValidationError as e:
                errors.extend(adjust_validation_error_location(e, location))

        if errors:
            raise ValidationError.from_exception_data(cls.__name__, errors)
        return values


def adjust_validation_error_location(
    e: ValidationError, location: tuple
) -> list[ErrorDetails]:
    errors = []
    for error in e.errors():
        error["loc"] = location + error["loc"]
        errors.append(error)
    return errors



from typing import Any, Literal

from polars import DataFrame
from pydantic import BaseModel

from ..interface import SourceConfig, SourceConfigArguments


class StaticConfigArguments(SourceConfigArguments):
    class ConnectionArguments(BaseModel):
        static: list[dict[str, Any]]

    type: Literal["Static"] = "Static"
    connection: ConnectionArguments


class StaticConfig(SourceConfig):
    """The connection configuration class used for static sources"""

    data: DataFrame

    def __init__(self, arguments: StaticConfigArguments):
        super().__init__(arguments)
        self.data = DataFrame(arguments.connection.static or [])



from typing import Annotated, Any

from pydantic import BaseModel, ConfigDict, Field


class ValidationBaseModel(BaseModel):
    model_config = ConfigDict(extra="forbid")
    location: Annotated[
        tuple[str, ...],
        Field(
            default=(),
            description="The location path of the model in the configuration file. Used to provide more detailed error messages.",
        ),
    ]

    @staticmethod
    def propagate_location(nested_cls_property, arguments: dict[str, Any]):
        """Propagates the location of the model to nested models. This should be used inside of a model_validator with mode='before' on every validation model that has nested validation models.

        Example:
        ```python
        class NestedModel(ValidationBaseModel):
          pass

        class MainModel(ValidationBaseModel):
          nested: NestedModel

          @model_validator(mode="before")
          def propagate_nested_location(cls, values: dict):
              propagate_location("nested", location)

              return values
        ```
        """
        nested_cls = arguments.get(nested_cls_property, None)
        if nested_cls is None:
            return
        location = arguments.get("location", ())
        nested_cls["location"] = location + (nested_cls_property,)



from typing import Annotated, Any, Dict, Literal, Optional, Union

from pydantic import ConfigDict, Field, model_validator

from ...errors import ConfigValueError
from ...validation import ValidationBaseModel
from ..interface import SourceSync, SourceSyncArguments


class RestResponse(ValidationBaseModel):
    length: Annotated[Optional[str], Field(default=None)]
    items: str


class BodyOptions(ValidationBaseModel):
    accepts_multiple: Annotated[bool, Field(default=False)]
    object_path: Annotated[str, Field(default="")]


class PaginationSettings(ValidationBaseModel):
    enabled: Annotated[bool, Field(default=False)]
    pageSize: Annotated[int, Field(default=10)]
    startPage: Annotated[int, Field(default=0)]
    pageSizeQuery: Annotated[str, Field(default="pageSize")]
    pageQuery: Annotated[str, Field(default="page")]
    totalQuery: Annotated[str, Field(default="page.count")]
    queryIn: Annotated[str, Field(default="query")]
    totalQueryIn: Annotated[str, Field(default="body")]


class RestRequest(ValidationBaseModel):
    path: Annotated[str, Field(default=None)]
    query: Annotated[Dict[str, Union[str, int]], Field(default={})]
    method: Annotated[
        Literal["GET", "POST", "PUT", "HEAD", "DELETE", "OPTIONS", "PATCH"],
        Field(default="GET"),
    ]
    pagination: Annotated[Optional[PaginationSettings], Field(default=None)]
    body_type: Annotated[str, Field(default="application/json")]
    body: Annotated[
        Optional[Union[dict[Union[str, int], Any], str]], Field(default=None)
    ]
    body_options: Annotated[Optional[BodyOptions], Field(default=None)]
    return_type: Annotated[str, Field(default="application/json")]
    response: Annotated[Optional[RestResponse], Field(default=None)]

    @model_validator(mode="after")
    def validate_body_type(cls, instance: "RestRequest"):
        if instance.body_type == "application/json" and isinstance(instance.body, str):
            raise ConfigValueError(
                "body",
                "Body must be of type dictionary when body_type is set to 'application/json'.",
                instance.location,
            )
        return instance


class RestSyncArguments(SourceSyncArguments):
    listRequest: RestRequest
    createRequest: Annotated[Optional[RestRequest], Field(default=None)]
    updateRequest: Annotated[Optional[RestRequest], Field(default=None)]
    deleteRequest: Annotated[Optional[RestRequest], Field(default=None)]
    type: Annotated[Literal["Rest"], Field(defauls="Rest")] = "Rest"

    @model_validator(mode="before")
    def propagate_nested_location(cls, values: dict):
        cls.propagate_location("listRequest", values)
        cls.propagate_location("createRequest", values)
        cls.propagate_location("updateRequest", values)
        cls.propagate_location("deleteRequest", values)

        return values


class RestSync(SourceSync):
    """The configuration class used for MySQL sources"""

    listRequest: RestRequest
    createRequest: Optional[RestRequest]
    updateRequest: Optional[RestRequest]
    deleteRequest: Optional[RestRequest]

    def __init__(self, arguments: RestSyncArguments):
        super().__init__(arguments)
        self.listRequest = arguments.listRequest
        self.createRequest = arguments.createRequest
        self.updateRequest = arguments.updateRequest
        self.deleteRequest = arguments.deleteRequest



from typing import Annotated, Literal, Optional

from pydantic import Field, model_validator

from ...errors import (
    ConfigValidationError,
    RequiredDestinationFieldError,
    RequiredSourceFieldError,
)
from ..interface import SourceSync, SourceSyncArguments


class PostgresSyncArguments(SourceSyncArguments):
    table: Annotated[Optional[str], Field(default=None)]
    query: Annotated[Optional[str], Field(default=None)]
    uniqueColumns: Annotated[list[str], Field(default=[])]
    skipColumns: Annotated[list[str], Field(default=[])]
    type: Annotated[Literal["Postgresql"], Field(defauls="Postgresql")] = "Postgresql"

    @model_validator(mode="after")
    def validate_as_source(cls, instance: "PostgresSyncArguments"):
        if instance.direction == "destination":
            return instance
        errors = []
        if not instance.table and not instance.query:
            errors.append(RequiredSourceFieldError("table|query", instance.location))

        if errors:
            raise ConfigValidationError(errors)
        return instance

    @model_validator(mode="after")
    def validate_as_destination(cls, instance: "PostgresSyncArguments"):
        if instance.direction == "source":
            return instance
        errors = []
        if not instance.table:
            errors.append(RequiredDestinationFieldError("table", instance.location))
        if not instance.uniqueColumns:
            errors.append(
                RequiredDestinationFieldError("uniqueColumns", instance.location)
            )

        if errors:
            raise ConfigValidationError(errors)

        return instance


class PostgresSync(SourceSync):
    """The configuration class used for Postgresql sources"""

    unique_columns: list[str] = None
    skip_columns: list[str] = None
    table: str = None
    query: str = None

    def __init__(self, arguments: PostgresSyncArguments):
        super().__init__(arguments)

        self.table = arguments.table
        self.query = arguments.query
        self.unique_columns = arguments.uniqueColumns
        self.skip_columns = arguments.skipColumns



from typing import Literal

import polars as pl

from .interface_config import SourceConfig, SourceConfigArguments
from .interface_sync import SourceSync, SourceSyncArguments

CASTABLE = (
    pl.Int8,
    pl.Int16,
    pl.Int32,
    pl.Int64,
    pl.UInt8,
    pl.UInt16,
    pl.UInt32,
    pl.UInt64,
    pl.Float32,
    pl.Float64,
    pl.Boolean,
    pl.Utf8,
    pl.Object,
    # Does not quite work
    # pl.Binary,
)


class SourceInterface:
    ConfigClass = SourceConfig
    ConfigArgumentsClass = SourceConfigArguments
    SyncClass = SourceSync
    SyncArgumentsClass = SourceSyncArguments

    """ Abstract interface for a connection to a data source """
    connection = None
    connection_settings_arguments = None
    connection_settings = None
    source_configuration_arguments = None
    source_configuration = None

    def __init__(self, source: dict) -> None:
        """Initiates a source class

        Args:
            config (dict): Coniguration for the source data (columns, IDs, etc.)
            connection (dict):
                Configuration for the source connection (paths, credentials, etc.)
        """

        self.connection_settings_arguments = self.ConfigArgumentsClass(**source)

        self.connection_settings = self.ConfigClass(self.connection_settings_arguments)

        self._configure()

    def set_sourceconfig(
        self,
        config: dict,
        direction: Literal["source", "destination"],
        name: str,
        location: tuple[str],
    ) -> None:
        self.source_configuration_arguments = self.SyncArgumentsClass(
            direction=direction, name=name, location=location, **config
        )
        self.source_configuration = self.SyncClass(self.source_configuration_arguments)

    def __enter__(self):
        self._connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._disconnect()

    def _configure(self):
        """
        Method responsible for configuring the connection to
        the dataset. Should be overridden in child class.
        Called during init.
        """
        raise NotImplementedError

    def _connect(self):
        """
        Method responsible for connecting to the dataset.
        Should be overridden in child class. Called during init.
        """
        raise NotImplementedError

    def _disconnect(self):
        """
        Method responsible for disconnecting from the dataset.
        Should be overridden in child class. Called during exit.
        """
        raise NotImplementedError

    def connect(self):
        self._connect()

    def disconnect(self):
        self._disconnect()

    def query(self, params=None):
        """Formats the query according to the column specification

        Args:
            df (DataFrame): The data from the data source
        """
        return self._query(params)

    def _query(self, params=None) -> pl.DataFrame:
        """Run a query on the source and return the results

        Args:
            params (mixed, optional): Parameters for the query. Defaults to None.

        Returns:
            pl.DataFrame: The results of the query
        """
        raise NotImplementedError

    def insert(self, data: pl.DataFrame) -> None:
        """Insert data into the source

        Args:
            data (pl.DataFrame): The data to insert
        """
        raise NotImplementedError

    def update(self, data: pl.DataFrame) -> None:
        """Update data in the source

        Args:
            data (pl.DataFrame): The data to update
        """
        raise NotImplementedError

    def delete(self, data: pl.DataFrame) -> None:
        """Delete data from the source

        Args:
            data (pl.DataFrame): The data to delete
        """
        raise NotImplementedError



from typing import Annotated, Literal, Optional

from pydantic import Field, model_validator

from ...errors import ConfigValidationError, ConfigValueError
from ...validation import ValidationBaseModel
from ..interface import SourceConfig, SourceConfigArguments, SourceConnectionArguments


class PostgresSettingsArguments(ValidationBaseModel):
    connection_string: Annotated[Optional[str], Field(default=None)]
    username: Annotated[Optional[str], Field(min_length=1, default=None)]
    password: Annotated[Optional[str], Field(min_length=1, default=None)]
    host: Annotated[Optional[str], Field(min_length=1, default=None)]
    port: Annotated[Optional[str], Field(min_length=1, default=None)]
    database: Annotated[Optional[str], Field(min_length=1, default=None)]

    @model_validator(mode="after")
    def validate_connection_string(
        cls, instance: "PostgresConfigArguments.ConnectionArguments"
    ):
        if instance.connection_string:
            return instance

        errors = []
        fields = ["username", "password", "host", "port", "database"]
        for field in fields:
            if getattr(instance, field) is None:
                errors.append(
                    ConfigValueError(
                        field,
                        f"'{field}' is required when 'connection_string' is not provided",
                        instance.location,
                    )
                )

        if errors:
            raise ConfigValidationError(errors)
        return instance


class PostgresConnectionArguments(SourceConnectionArguments):
    settings: PostgresSettingsArguments

    @model_validator(mode="before")
    def propagate_nested_location(cls, values: "PostgresConnectionArguments"):
        cls.propagate_location("settings", values)
        return values


class PostgresConfigArguments(SourceConfigArguments):
    type: Annotated[Literal["Postgresql"], Field(default="Postgresql")] = "Postgresql"
    connection: PostgresConnectionArguments


class PostgresConfig(SourceConfig):
    """The connection configuration class used for Postgresql sources"""

    connection_string: str

    def __init__(self, arguments: PostgresConfigArguments):
        super().__init__(arguments)
        self.connection_string = arguments.connection.settings.connection_string

        if not self.connection_string:
            self.connection_string = f"postgresql://{arguments.connection.settings.username}:{arguments.connection.settings.password}@{arguments.connection.settings.host}:{arguments.connection.settings.port}/{arguments.connection.settings.database}"



from typing import Annotated, Any, Literal, Optional

from pydantic import Field

from ..interface import SourceSync, SourceSyncArguments


class MongodbSyncArguments(SourceSyncArguments):
    collection: str
    filter: Annotated[Optional[dict[str, Any]], Field(default={})]
    projection: Annotated[Optional[dict[str, int]], Field(default={})]
    uniqueFields: Annotated[
        Optional[list[str]],
        Field(default=[], description="Required when used as a destination"),
    ]
    type: Annotated[Literal["Mongodb"], Field(default="Mongodb")] = "Mongodb"


class MongodbSync(SourceSync):
    """The configuration class used for MongoDB sources"""

    collection: Annotated[Optional[str], Field(default=None)]
    filter: Annotated[str, Field(default={})]
    projection: dict = None
    unique_fields: list[str] = None

    def __init__(self, arguments: MongodbSyncArguments):
        super().__init__(arguments)
        self.collection = arguments.collection
        self.filter = arguments.filter
        self.projection = arguments.projection
        self.unique_fields = arguments.uniqueFields



import os
from typing import Union

from .frames_schema import FramesTransformerSchemas
from .integer_schema import IntegerTransformerSchemas
from .itop_schema import ItopTransformerSchemas
from .misc_schema import MiscTransformerSchemas
from .regex_schema import RegexTransformerSchemas
from .strings_schema import StringTransformerSchemas
from .structs_schema import StructTransformerSchemas

TransformerSchemas = Union[
    StringTransformerSchemas,
    StructTransformerSchemas,
    RegexTransformerSchemas,
    MiscTransformerSchemas,
    ItopTransformerSchemas,
    IntegerTransformerSchemas,
    FramesTransformerSchemas,
]

# Automatic import of all transformers
cDir = os.path.dirname(__file__)
for file in os.listdir(cDir):
    if file.endswith(".py") and not file.startswith("__") and file != "interface.py":
        __import__(f"{__name__}.{file[:-3]}", fromlist=["*"])



from typing import Annotated, Literal, Optional

from pydantic import Field, model_validator
from pyodbc import DatabaseError, drivers

from ...errors import ConfigValidationError, ConfigValueError
from ...validation import ValidationBaseModel
from ..interface import SourceConfig, SourceConfigArguments, SourceConnectionArguments


class SqlserverSettingsArguments(ValidationBaseModel):
    connection_string: Annotated[
        Optional[str],
        Field(
            default=None,
            description="Required if individual connection settings are not provided",
        ),
    ]
    username: Annotated[
        Optional[str],
        Field(
            default=None, description="Required if connection string is not provided"
        ),
    ]
    password: Annotated[
        Optional[str],
        Field(
            default=None, description="Required if connection string is not provided"
        ),
    ]
    host: Annotated[
        Optional[str],
        Field(
            default=None, description="Required if connection string is not provided"
        ),
    ]
    port: Annotated[
        Optional[str],
        Field(
            default=None, description="Required if connection string is not provided"
        ),
    ]
    database: Annotated[
        Optional[str],
        Field(
            default=None, description="Required if connection string is not provided"
        ),
    ]

    @model_validator(mode="after")
    def validate_connection_settings(cls, instance: "SqlserverConnectionArguments"):
        if instance.connection_string:
            return instance

        errors = []
        required_fields = ["username", "password", "host", "port", "database"]
        for field in required_fields:
            if not getattr(instance, field):
                errors.append(
                    ConfigValueError(
                        field,
                        f"Field '{field}' is required when 'connection_string' is not provided",
                        instance.location,
                    )
                )
        if errors:
            raise ConfigValidationError(errors)

        return instance


class SqlserverConnectionArguments(SourceConnectionArguments):
    settings: SqlserverSettingsArguments

    @model_validator(mode="before")
    def propagate_nested_location(cls, arguments: "SqlserverConnectionArguments"):
        cls.propagate_location("settings", arguments)
        return arguments


class SqlserverConfigArguments(SourceConfigArguments):
    connection: SqlserverConnectionArguments
    type: Annotated[Literal["Sqlserver"], Field(default="Sqlserver")] = "Sqlserver"


class SqlserverConfig(SourceConfig):
    """The connection configuration class used for SQLServer sources"""

    connection_string: str

    def append_sqlserver_driver(self, connection_string: str):
        if "driver" in connection_string:
            return connection_string

        sql_drivers = [driver for driver in drivers() if "SQL Server" in driver]
        no_drivers_available = len(sql_drivers) == 0
        if no_drivers_available:
            raise DatabaseError(
                "No ODBC drivers found for SQL Server/pyodbc. "
                "Please make sure at least one is installed"
            )

        print(
            f"\nFound driver '{sql_drivers[0]}', beetl is automatically appending it to the connection string for you. If you want to use a different driver, please add it manually."
        )
        join_char = "&" if "?" in connection_string else "?"
        return f"{connection_string}{join_char}driver={sql_drivers[0]}"

    def __init__(self, arguments: SqlserverConfigArguments):
        super().__init__(arguments)
        if arguments.connection.settings.connection_string:
            self.connection_string = self.append_sqlserver_driver(
                arguments.connection.settings.connection_string
            )
            return

        composed_connection_string = "mssql://"
        f"{arguments.connection.settings.username}:{arguments.connection.settings.password}"
        f"@{arguments.connection.settings.host}:{arguments.connection.settings.port}/{arguments.connection.settings.database}"

        self.connection_string = self.append_sqlserver_driver(
            composed_connection_string
        )



import copy
import json
import os

from ...sources import Sources
from ...transformers.interface import TransformerConfiguration
from ..config_base import BeetlConfig, ComparisonColumn, SyncConfiguration
from .v1_schema import BeetlConfigSchemaV1


class BeetlConfigV1(BeetlConfig):
    """Version 1 of the Beetl Configuration Interface"""

    @staticmethod
    def generate_jsonschema() -> str:
        """Generates the jsonschema for the BeetlConfigV1 configuration."""
        return json.dumps(BeetlConfigSchemaV1.model_json_schema())

    @staticmethod
    def validate_config(config: dict):
        """Performs static and dynamic validation of the user supplied configuration.

        Args:
            config (dict): The user supplied configuration

        Raises:
            ValidationError: If the configuration is invalid when compared against the static jsonschema
            ConfigValidationError: If the configuration is invalid when compared against the dynamic beetl validation rules
        """
        BeetlConfigSchemaV1(**config)

    def __init__(self, config: dict):
        self.sources, self.sync_list = {}, []

        if not config.get("sources", []):
            config["sources"] = self.read_sources_from_env(config)

        BeetlConfigV1.validate_config(config)

        self.sources = self.initialize_sources(config)

        if len(config.get("sync", "")) == 0:
            raise Exception("The configuration file is missing the 'sync' section.")

        for sync_index, sync in enumerate(config["sync"]):
            location = ("sync", str(sync_index))
            if not self.sources.get(sync["source"], False) or not self.sources.get(
                sync["destination"], False
            ):
                raise Exception(
                    "One of the source/destination names in "
                    "the sync section does not match a source name "
                    "in the sources section."
                    "Please check your configuration.",
                )

            source_name = sync["source"]
            temp_source = copy.deepcopy(self.sources[source_name])
            temp_source.set_sourceconfig(
                sync["sourceConfig"],
                direction="source",
                name=source_name,
                location=(*location, "sourceConfig"),
            )

            destination_name = sync["destination"]
            temp_destination = copy.deepcopy(self.sources[destination_name])
            temp_destination.set_sourceconfig(
                sync["destinationConfig"],
                direction="destination",
                name=destination_name,
                location=(*location, "destinationConfig"),
            )

            comparisonColumnsConf = sync.get("comparisonColumns", None)
            if not comparisonColumnsConf:
                raise Exception(
                    "The sync configuration is missing the 'comparisonColumns' key."
                )
            if type(comparisonColumnsConf) is list:
                try:
                    comparisonColumns = [
                        ComparisonColumn(**args) for args in sync["comparisonColumns"]
                    ]
                except TypeError as e:
                    raise Exception(
                        "When passing the comparisonColumns as a list it must be a list of dictionaries containing the mandatory keys 'name', 'type', and optionally 'unique'."
                    ) from e
            elif type(comparisonColumnsConf) is dict:
                try:
                    comparisonColumns = []
                    unique_column_name, _ = list(comparisonColumnsConf.items())[0]
                    for key, value in comparisonColumnsConf.items():
                        comparisonColumns.append(
                            ComparisonColumn(key, value, key == unique_column_name)
                        )

                except Exception as e:
                    raise Exception(
                        "The comparisonColumns must be a list of dictionaries containing the mandatory keys 'name', 'type', and optionally 'unique'."
                    ) from e
            else:
                raise Exception(
                    "The comparisonColumns must be a list of dictionaries containing the mandatory keys 'name', 'type', and optionally 'unique'. Or a dictionary with the column names as key and types as values, where the first key is treated as the unique column."
                )

            syncConfig = SyncConfiguration(
                name=sync.get("name", ""),
                source=temp_source,
                sourceConfig=sync["sourceConfig"],
                destination=temp_destination,
                destinationConfig=sync["destinationConfig"],
                comparisonColumns=comparisonColumns,
            )

            temp_source = temp_destination = None

            if sync.get("sourceTransformer", None) is not None:
                syncConfig.sourceTransformer = TransformerConfiguration(
                    "source", sync["sourceTransformer"], {}
                )

            if sync.get("sourceTransformers", None) is not None:
                syncConfig.sourceTransformers = []
                for transformer in sync["sourceTransformers"]:
                    syncConfig.sourceTransformers.append(
                        TransformerConfiguration(
                            transformer.get("transformer"),
                            transformer.get("config", None),
                        )
                    )

            if sync.get("destinationTransformers", None) is not None:
                syncConfig.destinationTransformers = []
                for transformer in sync["destinationTransformers"]:
                    syncConfig.destinationTransformers.append(
                        TransformerConfiguration(
                            transformer.get("transformer"),
                            transformer.get("config", None),
                        )
                    )

            if sync.get("insertionTransformers", None) is not None:
                syncConfig.insertionTransformers = []
                for transformer in sync["insertionTransformers"]:
                    syncConfig.insertionTransformers.append(
                        TransformerConfiguration(
                            transformer.get("transformer"),
                            transformer.get("config", None),
                            transformer.get("include_sync", False),
                        )
                    )

            if sync.get("deletionTransformers", None) is not None:
                syncConfig.deletionTransformers = []
                for transformer in sync["deletionTransformers"]:
                    syncConfig.deletionTransformers.append(
                        TransformerConfiguration(
                            transformer.get("transformer"),
                            transformer.get("config", None),
                            transformer.get("include_sync", False),
                        )
                    )

            self.sync_list.append(syncConfig)

    def initialize_sources(self, config):
        sources = {}
        for source_index, source in enumerate(config["sources"]):
            name = source["name"]
            source_type = source["type"]
            registrated_source = Sources.sources.get(source_type, None)
            if registrated_source is None:
                raise Exception(
                    f"The source type '{source_type}' used in source '{name}' is not registrated in available sources"
                )

            location = ("sources", str(source_index))
            source_config_with_location = {**source, "location": location}
            sources[name] = registrated_source.cls(source_config_with_location)
        return sources

    def read_sources_from_env(self, config):
        not_using_env_config = config.get("sourcesFromEnv", "") in [None, ""]
        if not_using_env_config:
            return

        sourcesFromEnv = config.pop("sourcesFromEnv")
        envJson = os.environ.get(sourcesFromEnv)
        try:
            envConfig = json.loads(envJson)
        except Exception as e:
            raise Exception(
                "The environment variable specified in 'sourcesFromEnv' "
                "is not a valid JSON string or does not exist."
            ) from e

        return envConfig



from .interface_config import (
    SourceConfig,
    SourceConfigArguments,
    SourceConnectionArguments,
)
from .interface_source import CASTABLE, SourceInterface
from .interface_sync import SourceSync, SourceSyncArguments



from time import perf_counter
from typing import List, Union, get_args

import polars as pl
from tabulate import tabulate

from .compare.compare import Difftool
from .comparison_result import ComparisonResult
from .config import BeetlConfig, ComparisonColumn, SyncConfiguration
from .result import Result, SyncResult
from .sources import CASTABLE
from .transformers.interface import TransformerConfiguration

BENCHMARK = []


class Beetl:
    """The main class for BeETL. This class is responsible for orchestrating the ETL process."""

    config: Union[BeetlConfig, None] = None
    """Holds the BeETL Configuration"""

    def __init__(self, config: BeetlConfig):
        self.config = config

    @classmethod
    def from_yaml(cls, path: str, encoding: str = "utf-8") -> "Beetl":
        """Creates a Beetl instance from a YAML-formatted configuration file

        Args:
            path (str): The path to the file
            encoding (str, optional): The charset for the file. Defaults to 'utf-8'.

        Returns:
            Beetl: An instance of Beetl
        """
        return cls(BeetlConfig.from_yaml_file(path, encoding))

    @classmethod
    def from_json(cls, path: str, encoding: str = "utf-8") -> "Beetl":
        """Creates a Beetl instance from a JSON-formatted configuration file

        Args:
            path (str): The path to the file
            encoding (str, optional): The charset for the file. Defaults to 'utf-8'.

        Returns:
            Beetl: _description_
        """
        return cls(BeetlConfig.from_json_file(path, encoding))

    @staticmethod
    def compare_datasets(
        source: pl.DataFrame,
        destination: pl.DataFrame,
        keys: List[str] = ["id"],
        columns: List[ComparisonColumn] = [],
    ) -> tuple[pl.DataFrame, pl.DataFrame, pl.DataFrame]:
        """
        This function uses polars DataFrames to quickly compare two datasets and return the differences.

        Polars is roughly 10-50x faster than Pandas for this task, but this can vary depending on the dataset.

        Args:
            source (pl.DataFrame): The source dataset
            destination (pl.DataFrame): The destination dataset
            keys (List[str], optional): Unique keys in the dataset. Defaults to ["id"].
            columns (List[str], optional): List of columns to compare for differences. Defaults to [].

        Returns:
            List[pl.DataFrame]: A list of Polars Dataframes (Insert, Update, Delete) containing the differences between source and destination
        """

        if isinstance(source, get_args(Union[list, set, tuple])):
            source = pl.DataFrame(source)

        if isinstance(destination, get_args(Union[list, set, tuple])):
            destination = pl.DataFrame(destination)

        source = Beetl._initialize_columns_if_empty(source, columns)
        source = Beetl._cast_columns_to_types(source, columns)
        destination = Beetl._initialize_columns_if_empty(destination, columns)
        destination = Beetl._cast_columns_to_types(destination, columns)
        column_names = [col.name for col in columns]

        for column in column_names:

            if column not in source.columns and column not in destination.columns:
                raise Exception(
                    f"Column {column} does not exist in any of the datasets"
                )

            if column not in source.columns and len(source) > 0:
                source = source.with_columns(
                    pl.lit(None).alias(column).cast(destination[column].dtype)
                )

            if column not in destination.columns and len(destination) > 0:
                destination = destination.with_columns(
                    pl.lit(None).alias(column).cast(source[column].dtype)
                )

        columns_contain_list_column = any(
            [
                True
                for name in column_names
                if type(source.get_column(name).dtype) == pl.List
            ]
        )
        if columns_contain_list_column:
            raise ValueError(
                "Beetl does not support comparing list columns, please remove them from the sync.comparisonColumns field. If you didn't specify any non unique columns, Beetl will compare all columns by default. Please specify the columns you want to compare in the sync.comparisonColumns field."
            )

        # If source is empty, delete all in destination
        if len(source) == 0:
            return source, source, destination

        # If destination is empty, create all from source
        if len(destination) == 0:
            return (source, destination, destination)
        try:
            # Get rows that only exist in source (Creates)
            create = source.join(destination, on=keys, how="anti")

            # Get rows that exist in both and have differing values (Updates)
            update = source.join(destination, on=keys, how="semi").join(
                destination, on=column_names, how="anti", join_nulls=True
            )

            # Get rows that only exist in destination (Deletes)
            delete = destination.join(source, on=keys, how="anti")

        except pl.ColumnNotFoundError as e:
            raise ValueError(
                "One or more comparison columns do not exist \n"
                f"Source columns: {','.join(source.columns)} \n"
                f"Destination columns: {','.join(destination.columns)} \n"
                f"Comparison columns: {','.join(column_names)} \n"
            ) from e

        try:
            comparison_results = (
                create.select(source.columns),
                update.select(source.columns),
                delete.select(destination.columns),
            )
        except Exception:
            raise Exception(
                "Could not create comparison results. Most likely due to a mismatch in column names between source and destination."
            )

        return comparison_results

    @staticmethod
    def _initialize_columns_if_empty(source, columns):
        if len(source) == 0 and source.width == 0:
            for col in columns:
                source = source.with_columns(pl.Series(col.name, dtype=col.type))
        return source

    @staticmethod
    def _cast_columns_to_types(source, columns):
        for col in columns:
            if col.name in source.columns and col.type in CASTABLE:
                source = source.with_columns(pl.col(col.name).cast(col.type))
        return source

    def benchmark(self, text: str) -> None:
        """Inserts a benchmark into the log"""
        BENCHMARK.append({"text": text, "perf": perf_counter()})

        if len(BENCHMARK) > 1:
            print(
                BENCHMARK[-1]["text"]
                + ": "
                + str(round(BENCHMARK[-1]["perf"] - BENCHMARK[-2]["perf"], 5))
            )

    def runTransformers(
        self,
        source: pl.DataFrame,
        transformers: List[TransformerConfiguration],
        sync: SyncConfiguration,
    ) -> pl.DataFrame:
        transformed = source.clone()

        if transformers is not None and len(transformers) > 0:
            for transformer in transformers:
                if transformer.include_sync:
                    transformed = transformer.transform(transformed, sync=sync)
                    continue

                transformed = transformer.transform(transformed)

        return transformed

    def sync(
        self, dry_run: bool = False, generate_update_diff: bool = False
    ) -> Union[Result, List[ComparisonResult]]:
        """Executes the ETL process.

        Args:
            dry_run (bool, optional): If set to true, the function will not execute any database operations and will return the dataframes that would have been used in the operations. Defaults to False.
            generate_update_diff (bool, optional): If set to true, the function will return the differences between the source and destination update dataframes. Defaults to False.

        Returns:
            ComparisonResult: If the argument dry_run was passed as true, returns a list of ComparisionResult objects that contain the create, update and delete dataframes that would have been used by the destination if the dry_run argument was false.
            Result: A Result object containing the amount of inserts, updates and deletes for each sync
            list[tuple[list[DataFrame], set[str]]]: If generate_update_diff is set to true, returns a list of tuples containing the differences between the source and destination update dataframes and a set with the keys that have differences across the whole dataframe.

        The following steps will be performed:

        1. Load source and destination data

        2. Format source data to be compatible with destination through field and source transformers

        3. Compare source and destination data with compare_datasets

        4. Execute the respective insert, update and delete queries

        """
        dry_run_results = []
        generate_update_diff_results = []
        self.benchmark("Starting sync and retrieving source data")
        allAmounts = []
        for i, sync in enumerate(self.config.sync_list, 1):
            start = perf_counter()
            if sync.name != "":
                print(f"Starting sync: {sync.name}")
            else:
                print(f"Starting sync {i}")
            sync.source.connect()
            source_data = sync.source.query(sync.sourceConfig)
            sync.source.disconnect()
            self.benchmark("Finished data retrieval from source")

            sync.destination.connect()
            destination_data = sync.destination.query(sync.destinationConfig)
            self.benchmark("Finished data retrieval from destination")

            self.benchmark("Starting source data transformation")
            transformedSource = self.runTransformers(
                source_data, sync.sourceTransformers, sync
            )
            self.benchmark(
                "Finished source data transformation, starting destination transformation"
            )
            transformedDestination = self.runTransformers(
                destination_data, sync.destinationTransformers, sync
            )

            self.benchmark("Finished data transformation before comparison")

            self.benchmark("Starting comparison")
            unique_columns = [
                column.name for column in sync.comparisonColumns if column.unique
            ]
            if len(unique_columns) == 0:
                raise ValueError(
                    "You need to specify at least one unique column in the sync.comparisonColumns field"
                )
            create, update, delete = self.compare_datasets(
                transformedSource,
                transformedDestination,
                unique_columns,
                sync.comparisonColumns,
            )
            self.benchmark("Successfully extracted operations from dataset")

            load_and_compare = perf_counter() - start
            print(f"Load and compare took {load_and_compare} seconds")

            amount = {}

            print(
                f"Insert: {len(create)}, Update: {len(update)}, Delete: {len(delete)}"
            )

            if generate_update_diff:
                generate_update_diff_results.append(
                    Difftool.diff_update(
                        update,
                        transformedDestination,
                        [
                            column.name
                            for column in sync.comparisonColumns
                            if not column.unique
                        ],
                        unique_columns,
                    )
                )
                continue

            if dry_run:
                dry_run_results.append(
                    ComparisonResult(
                        self.runTransformers(create, sync.insertionTransformers, sync),
                        self.runTransformers(update, sync.insertionTransformers, sync),
                        self.runTransformers(delete, sync.deletionTransformers, sync),
                    )
                )
                sync.destination.disconnect()
                continue

            self.benchmark("Starting database operations")
            self.benchmark("Starting deletes")
            amount["deletes"] = 0
            if len(delete):
                amount["deletes"] = sync.destination.delete(
                    self.runTransformers(delete, sync.deletionTransformers, sync)
                )

            self.benchmark("Finished deletes, starting inserts")
            amount["inserts"] = 0
            if len(create):
                amount["inserts"] = sync.destination.insert(
                    self.runTransformers(create, sync.insertionTransformers, sync)
                )

            self.benchmark("Finished inserts, starting updates")

            amount["updates"] = 0
            if len(update):
                amount["updates"] = sync.destination.update(
                    self.runTransformers(update, sync.insertionTransformers, sync)
                )

            self.benchmark("Finished updates, sync finished")

            print("Inserted: " + str(amount["inserts"]))
            print("Updated: " + str(amount["updates"]))
            print("Deleted: " + str(amount["deletes"]))

            allAmounts.append(
                [sync.name, *[amount["inserts"], amount["updates"], amount["deletes"]]]
            )

            sync.destination.disconnect()

        if generate_update_diff:
            return generate_update_diff_results

        if dry_run:
            return dry_run_results

        print(
            "\r\n\r\n"
            + tabulate(allAmounts, headers=["Sync", "Inserts", "Updates", "Deletes"])
        )

        return SyncResult(allAmounts)



from typing import Annotated, Literal, Optional, Union

from pydantic import ConfigDict, Field, model_validator

from ...errors import ConfigValidationError, ConfigValueError
from ...sources.interface import (
    SourceConfig,
    SourceConfigArguments,
    SourceConnectionArguments,
)
from ...validation import ValidationBaseModel


class Oauth2Arguments(ValidationBaseModel):
    model_config = ConfigDict(extra="forbid")

    flow: Annotated[str, Field(default="")]
    token_url: Annotated[str, Field(default="")]
    authorization_url: Annotated[str, Field(default="")]
    token_body: Annotated[dict, Field(default={})]
    token_path: Annotated[str, Field(default="access_token")]
    valid_to_path: Annotated[str, Field(default="expires_on")]


class RestAuthenticationArguments(ValidationBaseModel):
    model_config = ConfigDict(extra="forbid")

    basic: Annotated[Optional[bool], Field(default=False)]
    basic_user: Annotated[Optional[str], Field(default=None)]
    basic_pass: Annotated[Optional[str], Field(default=None)]
    bearer: Annotated[Optional[bool], Field(default=False)]
    bearer_prefix: Annotated[str, Field(default="Bearer")]
    bearer_token: Annotated[Optional[str], Field(default=None)]
    oauth2: Annotated[Optional[bool], Field(default=False)]
    oauth2_settings: Annotated[Optional[Oauth2Arguments], Field(default=None)]

    @model_validator(mode="before")
    def propagate_nested_location(cls, values: dict):
        cls.propagate_location("oauth2_settings", values)
        return values

    @model_validator(mode="after")
    def validate_authentication(cls, instance: "RestAuthenticationArguments"):
        if not any([instance.basic, instance.bearer, instance.oauth2]):
            raise ValueError("At least one authentication method must be set")
        if sum([instance.basic, instance.bearer, instance.oauth2]) > 1:
            raise ValueError("Only one authentication method can be set")
        return instance


class RestSettingsArguments(ValidationBaseModel):
    model_config = ConfigDict(extra="forbid")
    base_url: Annotated[str, Field(min_length=1)]
    ignore_certificates: Annotated[bool, Field(default=False)]
    authentication: Annotated[
        Optional[RestAuthenticationArguments], Field(default=None)
    ]

    # TODO: These should be moved to the sync config as they are dependant on wether the source is a destination or not
    soft_delete: Annotated[bool, Field(default=False)]
    deleted_field: Annotated[str, Field(default="deleted")]
    deleted_value: Annotated[Union[None, str, int], Field(default=None)]

    @model_validator(mode="before")
    def propagate_nested_location(cls, values: dict):
        cls.propagate_location("authentication", values)
        return values

    @model_validator(mode="after")
    def validate_soft_delete(
        cls,
        instance: "RestConfigArguments.ConnectionArguments.SettingsArguments",
    ):
        if not instance.soft_delete:
            return instance

        errors = []
        required_fields = ["deleted_field", "deleted_value"]
        for field in required_fields:
            if not getattr(instance, field):
                errors.append(
                    ConfigValueError(
                        field,
                        f"Field '{field}' is required when using 'soft_delete' is set to True",
                        instance.location,
                    )
                )
        if errors:
            raise ConfigValidationError(errors)

        return instance


class ConnectionArguments(SourceConnectionArguments):
    settings: RestSettingsArguments

    @model_validator(mode="before")
    def propagate_nested_location(cls, values: dict):
        cls.propagate_location("settings", values)
        return values


class RestConfigArguments(SourceConfigArguments):
    connection: ConnectionArguments
    type: Annotated[Literal["Rest"], Field(default="Rest")] = "Rest"


class RestConfig(SourceConfig):
    """The connection configuration class used for MySQL sources"""

    base_url: str
    ignore_certificates: bool
    authentication: Optional[RestAuthenticationArguments]
    soft_delete: bool
    deleted_field: str
    deleted_value: Union[None, str, int]

    def __init__(self, arguments: RestConfigArguments):
        super().__init__(arguments)

        settings = arguments.connection.settings

        self.base_url = settings.base_url
        self.authentication = settings.authentication or None
        self.ignore_certificates = settings.ignore_certificates
        self.soft_delete = settings.soft_delete
        self.deleted_field = settings.deleted_field
        self.deleted_value = settings.deleted_value



from typing import Annotated, Optional

from pydantic import Field, model_validator

from ...validation import ValidationBaseModel
from ..interface import SourceConfig, SourceConfigArguments, SourceConnectionArguments


class MysqlSettingsArguments(ValidationBaseModel):
    connection_string: Annotated[Optional[str], Field(default=None)]
    username: Annotated[Optional[str], Field(default=None)]
    password: Annotated[Optional[str], Field(default=None)]
    host: Annotated[Optional[str], Field(default=None)]
    port: Annotated[Optional[str], Field(default=None)]
    database: Annotated[Optional[str], Field(default=None)]

    @model_validator(mode="after")
    def validate_connection_string_or_components(
        cls, instance: "MysqlSettingsArguments"
    ):
        connection_string_is_not_present = not instance.connection_string
        connection_string_components = [
            "host",
            "port",
            "username",
            "password",
            "database",
        ]
        if connection_string_is_not_present:
            dict = instance.model_dump()
            for component in connection_string_components:
                if not dict.get(component, None):
                    raise ValueError(
                        f"'{component}' is missing. {connection_string_components} are required if 'connection_string' is not provided"
                    )
        return instance


class MysqlConnectionArguments(SourceConnectionArguments):
    settings: MysqlSettingsArguments

    @model_validator(mode="before")
    def propagate_nested_location(cls, values: dict):
        cls.propagate_location("settings", values)
        return values


class MysqlConfigArguments(SourceConfigArguments):
    type: Annotated[str, Field(default="Mysql")] = "Mysql"
    connection: MysqlConnectionArguments


class MysqlConfig(SourceConfig):
    """The connection configuration class used for MySQL sources"""

    connection_string: str
    query: str = None
    table: str = None

    def __init__(self, arguments: MysqlConfigArguments):
        super().__init__(arguments)

        connection_string: str
        if arguments.connection.settings.connection_string:
            connection_string = arguments.connection.settings.connection_string
        if not arguments.connection.settings.connection_string:
            connection_string = "mysql+pymysql://"
            f"{arguments.connection.settings.username}:{arguments.connection.settings.password}"
            f"@{arguments.connection.settings.host}:{arguments.connection.settings.port}/{arguments.connection.settings.database}"

        self.connection_string = connection_string



from typing import Annotated, Literal, Optional, Union

from pydantic import Field, model_validator

from .interface import TransformerConfigBase, TransformerSchemaBase


class StringTransformerSchema:
    class StaticField(TransformerSchemaBase):
        class Config(TransformerConfigBase):
            field: Annotated[str, Field(min_length=1)]
            value: str

        transformer: Literal["strings.staticfield"]
        config: Config

    class SetDefaults(TransformerSchemaBase):
        class Config(TransformerConfigBase):
            inField: Annotated[str, Field(min_length=1)]
            defaultValue: str

        transformer: Literal["strings.set_default"]
        config: Config

    class Strip(TransformerSchemaBase):
        class Config(TransformerConfigBase):
            inField: Annotated[str, Field(min_length=1)]
            stripChars: str
            outField: Annotated[Optional[str], Field(default=None)]

        transformer: Literal["strings.strip"]
        config: Config

    class Lowercase(TransformerSchemaBase):
        class Config(TransformerConfigBase):
            inField: Annotated[str, Field(min_length=1)]
            outField: Annotated[Optional[str], Field(default=None)]
            inOutMap: Annotated[dict[str, str], Field(default={})]

        transformer: Literal["strings.lowercase"]
        config: Config

    class Uppercase(TransformerSchemaBase):
        class Config(TransformerConfigBase):
            outField: Annotated[Optional[str], Field(default=None)]
            inOutMap: Annotated[dict[str, str], Field(default={})]
            inField: Annotated[str, Field(default="")]

            @model_validator(mode="wrap")
            def model_validator(cls, values: dict, handler):
                model: "StringTransformerSchema.Uppercase.Config" = handler(values)

                in_out_map_missing_or_empty: bool = (
                    not model.inOutMap or len(model.inOutMap) == 0
                )
                if in_out_map_missing_or_empty:
                    if model.inField is None:
                        raise ValueError(
                            "inField is required when inOutMap is missing or empty"
                        )
                    if len(model.inField) == 0:
                        raise ValueError("inField must have a length greater than 0")
                return model

        transformer: Literal["strings.uppercase"]
        config: Config

    class MatchContains(TransformerSchemaBase):
        class Config(TransformerConfigBase):
            inField: Annotated[str, Field(min_length=1)]
            outField: Annotated[Optional[str], Field(default=None)]
            match: str

        transformer: Literal["strings.match_contains"]
        config: Config

    class Join(TransformerSchemaBase):
        class Config(TransformerConfigBase):
            inFields: Annotated[list[str], Field(min_length=1)]
            outField: Annotated[str, Field(min_length=1)]
            separator: Annotated[str, Field(default="")]

        transformer: Literal["strings.join"]
        config: Config

    class JoinListfield(TransformerSchemaBase):
        class Config(TransformerConfigBase):
            inField: Annotated[str, Field(min_length=1)]
            outField: Annotated[str, Field(min_length=1)]
            separator: Annotated[str, Field(default="")]

        transformer: Literal["strings.join_listfield"]
        config: Config

    class Split(TransformerSchemaBase):
        class Config(TransformerConfigBase):
            inField: Annotated[str, Field(min_length=1)]
            outFields: Annotated[list[str], Field(min_length=1)]
            separator: Annotated[str, Field(default="")]

        transformer: Literal["strings.split"]
        config: Config

    class SplitIntoListfield(TransformerSchemaBase):
        class Config(TransformerConfigBase):
            inField: Annotated[str, Field(min_length=1)]
            outField: Annotated[Optional[str], Field(default=None)]
            separator: Annotated[str, Field(default="")]

        transformer: Literal["strings.split_into_listfield"]
        config: Config

    class Quote(TransformerSchemaBase):
        class Config(TransformerConfigBase):
            inField: Annotated[str, Field(min_length=1)]
            outField: Annotated[Optional[str], Field(default=None)]
            quote: Annotated[str, Field(default="'")]

        transformer: Literal["strings.quote"]
        config: Config

    class Replace(TransformerSchemaBase):
        class Config(TransformerConfigBase):
            inField: Annotated[str, Field(min_length=1)]
            outField: Annotated[Optional[str], Field(default=None)]
            search: str
            replace: str

        transformer: Literal["strings.replace"]
        config: Config

    class ReplaceAll(TransformerSchemaBase):
        class Config(TransformerConfigBase):
            inField: Annotated[str, Field(min_length=1)]
            outField: Annotated[Optional[str], Field(default=None)]
            search: str
            replace: str

        transformer: Literal["strings.replace_all"]
        config: Config

    class Substring(TransformerSchemaBase):
        class Config(TransformerConfigBase):
            inField: Annotated[str, Field(min_length=1)]
            outField: Annotated[Optional[str], Field(default=None)]
            start: Annotated[int, Field(default=0)]
            length: Annotated[Optional[int], Field(default=None)]

        transformer: Literal["strings.substring"]
        config: Config

    class AddPrefix(TransformerSchemaBase):
        class Config(TransformerConfigBase):
            inField: Annotated[str, Field(min_length=1)]
            outField: Annotated[Optional[str], Field(default=None)]
            prefix: Annotated[str, Field(min_length=1)]

        transformer: Literal["strings.add_prefix"]
        config: Config

    class Cast(TransformerSchemaBase):
        class Config(TransformerConfigBase):
            inField: Annotated[str, Field(min_length=1)]
            outField: Annotated[str, Field(default="")]

        transformer: Literal["strings.cast"]
        config: Config

    class ToObjectId(TransformerSchemaBase):
        class Config(TransformerConfigBase):
            inField: Annotated[str, Field(min_length=1)]

        transformer: Literal["strings.to_object_id"]
        config: Config

    class Format(TransformerSchemaBase):
        class Config(TransformerConfigBase):
            inField: Annotated[str, Field(min_length=1)]
            outField: Annotated[str, Field(default="")]
            format_string: Annotated[str, Field(default="{value}")]

        transformer: Literal["strings.format"]
        config: Config

    class Hash(TransformerSchemaBase):
        class Config(TransformerConfigBase):
            inField: Annotated[Optional[str], Field(default=None)]
            inFields: Annotated[list[str], Field(default=[])]
            outField: Annotated[str, Field(min_length=1)]
            hashWhen: Annotated[
                Literal["always", "any-value-is-populated", "all-values-are-populated"],
                Field(default="always"),
            ]

            @model_validator(mode="after")
            def validate_fields(cls, model: "StringTransformerSchema.Hash.Config"):
                if not model.inField and not model.inFields:
                    raise ValueError("Either inField or inFields must be provided")
                if len(model.inField or "") > 0 and len(model.inFields) > 0:
                    raise ValueError("Only one of inField or inFields can be provided")
                return model

        transformer: Literal["strings.hash"]
        config: Config


StringTransformerSchemas = Union[
    StringTransformerSchema.StaticField,
    StringTransformerSchema.SetDefaults,
    StringTransformerSchema.Strip,
    StringTransformerSchema.Lowercase,
    StringTransformerSchema.Uppercase,
    StringTransformerSchema.MatchContains,
    StringTransformerSchema.Join,
    StringTransformerSchema.JoinListfield,
    StringTransformerSchema.Split,
    StringTransformerSchema.SplitIntoListfield,
    StringTransformerSchema.Quote,
    StringTransformerSchema.Replace,
    StringTransformerSchema.ReplaceAll,
    StringTransformerSchema.Substring,
    StringTransformerSchema.AddPrefix,
    StringTransformerSchema.Cast,
    StringTransformerSchema.Hash,
    StringTransformerSchema.ToObjectId,
    StringTransformerSchema.Format,
    StringTransformerSchema.Hash,
]



from typing import Annotated, Literal, Optional

from pydantic import ConfigDict, Field, model_validator

from ...errors import ConfigValidationError, RequiredDestinationFieldError
from ..interface import SourceSync, SourceSyncArguments


class MysqlSyncArguments(SourceSyncArguments):
    model_config = ConfigDict(extra="forbid")

    table: Annotated[
        Optional[str],
        Field(
            default=None,
            description="Mandatory if configured as a destination, and if query isn't provided",
        ),
    ]
    query: Annotated[Optional[str], Field(default=None)]
    uniqueColumns: Annotated[list[str], Field(default=[])]
    skipColumns: Annotated[list[str], Field(default=[])]
    type: Annotated[Literal["Mysql"], Field(defauls="Mysql")] = "Mysql"

    @model_validator(mode="after")
    def validate_as_source(cls, instance: "MysqlSyncArguments"):
        if instance.direction == "destination":
            return instance

        if not instance.table and not instance.query:
            raise ValueError("'table' or 'query' is required when used as a source")

        return instance

    @model_validator(mode="after")
    def validate_as_destination(cls, instance: "MysqlSyncArguments"):
        if instance.direction == "source":
            return instance

        errors = []

        if not instance.table and not instance.query:
            errors.append(RequiredDestinationFieldError("table", instance.location))

        if not instance.uniqueColumns:
            errors.append(
                RequiredDestinationFieldError("uniqueColumns", instance.location)
            )

        if errors:
            raise ConfigValidationError(errors)

        return instance


class MysqlSync(SourceSync):
    """The configuration class used for MySQL sources"""

    unique_columns: list[str] = None
    skip_columns: list[str] = None
    table: str = None
    query: str = None

    def __init__(self, arguments: MysqlSyncArguments):
        super().__init__(arguments)

        self.table = arguments.table
        self.query = arguments.query
        self.unique_columns = arguments.uniqueColumns
        self.skip_columns = arguments.skipColumns



from polars import DataFrame, Object
from pymongo import DeleteOne, MongoClient, UpdateOne

from ..interface import SourceInterface
from ..registrated_source import register_source
from .mongodb_config import MongodbConfig, MongodbConfigArguments
from .mongodb_sync import MongodbSync, MongodbSyncArguments


@register_source("Mongodb")
class MongodbSource(SourceInterface):
    ConfigArgumentsClass = MongodbConfigArguments
    ConfigClass = MongodbConfig
    SyncArgumentsClass = MongodbSyncArguments
    SyncClass = MongodbSync

    """ A source for MongoDB data """

    def _configure(self):
        pass

    def _connect(self):
        pass

    def _disconnect(self):
        pass

    def _query(self, params=None) -> DataFrame:
        with MongoClient(self.connection_settings.connection_string) as client:
            db = client[self.connection_settings.database]
            collection = db[self.source_configuration.collection]
            polar = DataFrame(
                collection.find(
                    self.source_configuration.filter,
                    projection=self.source_configuration.projection,
                )
            )
            if "_id" in polar.columns and polar["_id"].dtype == Object:
                polar = polar.with_columns(
                    polar["_id"].map_elements(lambda oid: str(oid), return_dtype=str)
                )
            return polar

    def insert(self, data: DataFrame) -> int:
        with MongoClient(self.connection_settings.connection_string) as client:
            db = client[self.connection_settings.database]
            collection = db[self.source_configuration.collection]
            result = collection.insert_many(data.to_dicts())

        return len(result.inserted_ids)

    def update(self, data: DataFrame) -> int:
        self._validate_unique_fields()
        updates = []
        for row in data.to_dicts():
            filter = {
                field_name: row[field_name]
                for field_name in self.source_configuration.unique_fields
            }
            for field_name in self.source_configuration.unique_fields:
                del row[field_name]
            updates.append(UpdateOne(filter, {"$set": row}))

        if not len(updates):
            return 0

        with MongoClient(self.connection_settings.connection_string) as client:
            db = client[self.connection_settings.database]
            collection = db[self.source_configuration.collection]
            result = collection.bulk_write(updates)

        return result.modified_count

    def delete(self, data: DataFrame) -> int:
        self._validate_unique_fields()
        deletes = []
        for row in data.to_dicts():
            filter = {
                field_name: row[field_name]
                for field_name in self.source_configuration.unique_fields
            }
            deletes.append(DeleteOne(filter))

        with MongoClient(self.connection_settings.connection_string) as client:
            db = client[self.connection_settings.database]
            collection = db[self.source_configuration.collection]
            result = collection.bulk_write(deletes)

        return result.deleted_count

    def _validate_unique_fields(self):
        if not self.source_configuration.unique_fields:
            raise ValueError(
                "Unique fields are required for MongoDB when used as a destination"
            )



from typing import Annotated, Literal, Optional

from pydantic import BaseModel, ConfigDict, Field, model_validator

from ...validation import ValidationBaseModel
from ..interface import SourceConfig, SourceConfigArguments, SourceConnectionArguments


class MongodbSettingsArguments(ValidationBaseModel):
    connection_string: Annotated[Optional[str], Field(min_length=1, default=None)]
    host: Annotated[Optional[str], Field(min_length=1, default=None)]
    port: Annotated[Optional[str], Field(min_length=1, default=None)]
    username: Annotated[Optional[str], Field(min_length=1, default=None)]
    password: Annotated[Optional[str], Field(min_length=1, default=None)]
    database: Annotated[str, Field(min_length=1)]

    @model_validator(mode="after")
    def validate_connection_string_or_components(
        cls, instance: "MongodbSettingsArguments"
    ):
        connection_string_is_not_present = not instance.connection_string
        connection_string_components = ["host", "port", "username", "password"]
        if connection_string_is_not_present:
            dict = instance.model_dump()
            for component in connection_string_components:
                if not dict.get(component, None):
                    raise ValueError(
                        f"'{component}' is missing. {connection_string_components} are required if 'connection_string' is not provided"
                    )

        return instance


class MongodbConnectionArguments(SourceConnectionArguments):
    settings: MongodbSettingsArguments

    @model_validator(mode="before")
    def propagate_nested_location(cls, values: dict):
        cls.propagate_location("settings", values)
        return values


class MongodbConfigArguments(SourceConfigArguments):

    type: Annotated[Literal["Mongodb"], Field(default="Mongodb")] = "Mongodb"
    connection: MongodbConnectionArguments


class MongodbConfig(SourceConfig):
    """The connection configuration class used for MongoDB sources"""

    connection_string: str
    query: Optional[str] = None
    database: str

    def __init__(self, arguments: MongodbConfigArguments):
        super().__init__(arguments)

        connection_string: str
        if arguments.connection.settings.connection_string:
            connection_string = arguments.connection.settings.connection_string
        else:
            connection_string = f"mongodb://{arguments.connection.settings.username}:{arguments.connection.settings.password}@{arguments.connection.settings.host}:{arguments.connection.settings.port}/"

        self.connection_string = connection_string
        self.database = arguments.connection.settings.database



from typing import Annotated, Any, Literal, Optional, Union

from pydantic import Field

from .interface import TransformerConfigBase, TransformerSchemaBase


class StructTransformerSchema:
    class StaticField(TransformerSchemaBase):
        class Config(TransformerConfigBase):
            field: Annotated[str, Field(min_length=1)]
            value: Any
            only_add_if_missing: Optional[bool] = False

        transformer: Literal["structs.staticfield"]
        config: "StructTransformerSchema.StaticField.Config"

    class Jsonpath(TransformerSchemaBase):
        class Config(TransformerConfigBase):
            inField: Annotated[str, Field(min_length=1)]
            outField: Annotated[str, Field(min_length=1)]
            jsonPath: Annotated[str, Field(min_length=1)]
            defaultValue: Annotated[str, Field(default="")]

        transformer: Literal["structs.jsonpath"]
        config: "StructTransformerSchema.Jsonpath.Config"

    class ComposeStruct(TransformerSchemaBase):
        class Config(TransformerConfigBase):
            outField: Annotated[str, Field(min_length=1)]
            map: Annotated[dict[str, str], Field(min_length=1)]

        transformer: Literal["structs.compose_struct"]
        config: "StructTransformerSchema.ComposeStruct.Config"

    class ComposeListOfStructs(TransformerSchemaBase):
        class Config(TransformerConfigBase):
            outField: Annotated[str, Field(min_length=1)]
            map: Annotated[dict[str, str], Field(min_length=1)]

        transformer: Literal["structs.compose_list_of_structs"]
        config: "StructTransformerSchema.ComposeListOfStructs.Config"


StructTransformerSchemas = Union[
    StructTransformerSchema.StaticField,
    StructTransformerSchema.Jsonpath,
    StructTransformerSchema.ComposeStruct,
    StructTransformerSchema.ComposeListOfStructs,
]



import json
from typing import List

import polars as pl

from .interface import TransformerInterface, register_transformer_class


@register_transformer_class("frames")
class FrameTransformer(TransformerInterface):
    @staticmethod
    def filter(data: pl.DataFrame, filter: dict, reverse: bool = False):
        """Filter a DataFrame

        Args:
            data (pl.DataFrame): The DataFrame to filter
            filter (dict): The filter to apply

        Returns:
            pl.DataFrame: The filtered DataFrame
        """
        __class__._validate_fields(data.columns, [x for x in filter.keys()])

        for key, value in filter.items():
            if value is None and reverse:
                data = data.filter(data[key].is_not_null())
            elif value is None:
                data = data.filter(data[key].is_null())
            elif reverse:
                data = data.filter(data[key] != value)
            else:
                data = data.filter(data[key] == value)

        return data

    @staticmethod
    def conditional(
        data: pl.DataFrame,
        conditionField: str,
        ifTrue: str,
        ifFalse: str,
        targetField: str = "",
    ):
        """Apply a conditional to a DataFrame

        Args:
            data (pl.DataFrame): The DataFrame to apply the conditional to
            conditionField (str): The field to check
            targetField (str): The field to add the values to
            ifTrue (str): The value to set if the condition is true
            ifFalse (str): The value to set if the condition is false

        Returns:
            pl.DataFrame: The DataFrame with the conditional applied
        """
        if targetField == "":
            targetField = conditionField

        __class__._validate_fields(data.columns, [conditionField])

        data = data.with_columns(
            pl.when(data[conditionField] == True)
            .then(ifTrue)
            .otherwise(ifFalse)
            .alias(
                targetField
                if targetField is not None and targetField != ""
                else conditionField
            )
        )

        return data

    @staticmethod
    def rename_columns(data: pl.DataFrame, columns: List[dict]):
        """Rename columns in the dataset

        Args:
            data (pl.DataFrame): The dataset
            columns (List[dict]): A list of dicts (from: col1, to: col2), or a dict (Dict[string, string]) {from_column: to_column}


        Returns:
            pl.DataFrame: DataFrame with renamed columns
        """
        columns_to_validate = []
        if isinstance(columns, dict):
            columns_to_validate = list(columns.keys())
        else:
            columns_to_validate = [x["from"] for x in columns]

        __class__._validate_fields(data.columns, columns_to_validate)
        ncolumns = columns

        if isinstance(columns, dict):
            ncolumns = []
            for k, v in columns.items():
                ncolumns.append({"from": k, "to": v})

        for column in ncolumns:
            data = data.rename({column["from"]: column["to"]})

        return data

    @staticmethod
    def copy_columns(data: pl.DataFrame, columns: List[dict]):
        """Copies a given column to another column

        Args:
            data (pl.DataFrame): The dataset
            columns (List[dict]): A list of dicts (from: col1, to: col2)

        Returns:
            pl.DataFrame: DataFrame with renamed columns
        """
        __class__._validate_fields(data.columns, [x["from"] for x in columns])
        for column in columns:
            fromField = column["from"]
            toField = column["to"]
            data = data.with_columns(data[fromField].alias(toField))

        return data

    @staticmethod
    def drop_columns(data: pl.DataFrame, columns: List[str]) -> pl.DataFrame:
        """Drop columns from a DataFrame

        Args:
            data (pl.DataFrame): Drop columns from this dataframe
            columns (List[str]): The columns to dr op

        Returns:
            pl.DataFrame: The dataframe without the columns
        """
        __class__._validate_fields(data.columns, columns)
        return data.drop(columns)

    @staticmethod
    def project_columns(data: pl.DataFrame, columns: List[str]) -> pl.DataFrame:
        """Project the DataFrame to only include the specified columns

        Args:
            data (pl.DataFrame): DataFrame to project
            columns (List[str]): The columns to keep

        Returns:
            pl.DataFrame: The projected DataFrame
        """
        __class__._validate_fields(data.columns, columns)
        return data.select(columns)

    @staticmethod
    def distinct(data: pl.DataFrame, columns: List[str] = None) -> pl.DataFrame:
        """Only keep distinct rows in the dataframe. Optionally set a list of columns

        Args:
            data (pl.DataFrame): The DataFrame to get distinct values from
            columns (List[str]) (optional): The columns to get distinct values from

        Returns:
            pl.DataFrame: DataFrame with only the distinct values
        """
        if columns is not None and len(columns) > 0:
            __class__._validate_fields(data.columns, columns)

        return data.unique(
            subset=columns if columns is not None and len(columns) > 0 else None,
            maintain_order=True,
        )

    @staticmethod
    def extract_nested_rows(
        data: pl.DataFrame, iterField: str = "", fieldMap: dict = {}, colMap: dict = {}
    ):
        """
        Convert each list item in a list column to a row

        Args:
            data (pl.DataFrame): Multiply rows in this database
            inField (str): The list column to get values from
            outField (str): The column in the result to put the iterated value in
            column_path (str): The path in the list item to the value of the new column
            includeColumns (List[str]): Columns to include in the result
        """
        new_obj = []

        def extract_field(data: dict, path: str):
            splitPath = path.split(".")
            for segment in splitPath:
                try:
                    iSegment = int(segment)
                    if len(data) <= iSegment:
                        return ""

                    data = data[iSegment]
                except Exception:
                    try:
                        data = data[segment]
                    except Exception:
                        return ""

            return data

        for row in data.iter_rows(named=True):
            cData = row[iterField]
            if isinstance(cData, str):
                cData = json.loads(cData)

            for i in range(len(cData)):
                new_row = {}

                if fieldMap is None or len(fieldMap) == 0:
                    new_row["value"] = cData[i]

                if isinstance(cData[i], str):
                    if len(fieldMap) == 1:
                        new_row[list(fieldMap.values())[0]] = cData[i]
                else:
                    for field, path in fieldMap.items():
                        new_row[field] = extract_field(cData[i], path)

                for dst, src in colMap.items():
                    new_row[dst] = row[src]

                new_obj.append(new_row)

        return pl.DataFrame(new_obj)



from typing import Annotated, Literal, Optional

from pydantic import Field, model_validator

from ...errors import (
    ConfigValidationError,
    ConfigValueError,
    ForbiddenSourceFieldError,
    RequiredDestinationFieldError,
)
from ..interface import SourceSync, SourceSyncArguments


class SqlserverSyncArguments(SourceSyncArguments):
    type: Annotated[Literal["Sqlserver"], Field(defauls="Sqlserver")] = "Sqlserver"
    table: Annotated[
        Optional[str],
        Field(
            default=None,
            description="Sql server table name. Required when used as a destination and when query isn't provided.",
        ),
    ]
    query: Annotated[
        Optional[str],
        Field(
            default=None,
            description="SQL query. Defaults to all fields in the specified 'table'.",
        ),
    ]
    soft_delete: Annotated[
        Optional[bool],
        Field(
            default=None,
            description="Enables soft delete. Only allowed when used as a destination.",
        ),
    ]
    deleted_field: Annotated[
        Optional[str],
        Field(
            default=None,
            description="Field used for soft delete. Only allowed when used as a destination.",
        ),
    ]
    deleted_value: Annotated[
        Optional[str],
        Field(
            default=None,
            description="Value used for soft delete. Only allowed when used as a destination.",
        ),
    ]
    uniqueColumns: Annotated[
        list[str],
        Field(
            default=[],
            description="Defines what columns to use as unique identifiers. Only allowed when used as a destination.",
        ),
    ]
    skipColumns: Annotated[
        list[str],
        Field(
            default=[],
            description="Defines what columns to skip when inserting and updating. Only allowed when used as a destination.",
        ),
    ]

    @model_validator(mode="after")
    def validate_as_source(cls, instance: "SqlserverSyncArguments"):
        if instance.direction == "destination":
            return instance

        errors = []

        if not instance.query:
            if not instance.table:
                errors.append(
                    ConfigValueError(
                        "table",
                        "Table name is required when query isn't provided",
                        instance.location,
                    )
                )

        forbidden_fields = [
            "deleted_field",
            "deleted_value",
            "soft_delete",
            "uniqueColumns",
            "skipColumns",
        ]
        for field in forbidden_fields:
            if getattr(instance, field):
                errors.append(ForbiddenSourceFieldError(field, instance.location))

        if errors:
            raise ConfigValidationError(errors)

        return instance

    @model_validator(mode="after")
    def validate_as_destination(cls, instance: "SqlserverSyncArguments"):
        if instance.direction == "source":
            return instance

        errors = []
        required_fields = ["uniqueColumns", "table"]
        for field in required_fields:
            if not getattr(instance, field):
                errors.append(
                    RequiredDestinationFieldError(
                        field,
                        instance.location,
                    )
                )

        if instance.soft_delete:
            required_fields = ["deleted_field", "deleted_value"]
            for field in required_fields:
                if not getattr(instance, field):
                    errors.append(
                        ConfigValueError(
                            field,
                            f"Field '{field}' is required when using 'soft_delete' is set to True",
                            instance.location,
                        )
                    )

        if errors:
            raise ConfigValidationError(errors)

        return instance


class SqlserverSync(SourceSync):
    """The configuration class used for SQLServer sources"""

    table: str = None
    query: str = None
    soft_delete: bool = False
    deleted_field: str = None
    deleted_value: str = "'true'"
    unique_columns: list[str] = None
    skip_columns: list[str] = None

    def __init__(
        self,
        arguments: SqlserverSyncArguments,
    ):
        super().__init__(arguments)
        self.table = arguments.table
        self.query = (
            arguments.query if arguments.query else f"SELECT * FROM {arguments.table}"
        )
        self.soft_delete = arguments.soft_delete
        self.deleted_field = arguments.deleted_field
        self.deleted_value = arguments.deleted_value
        self.unique_columns = arguments.uniqueColumns
        self.skip_columns = arguments.skipColumns



from typing import Annotated, Literal, Optional, Union

from pydantic import Field

from .interface import TransformerConfigBase, TransformerSchemaBase


class RegexTransformerSchema:
    class Replace(TransformerSchemaBase):
        class Config(TransformerConfigBase):
            inField: Annotated[str, Field(min_length=1)]
            outField: Annotated[Optional[str], Field(default=None)]
            query: Annotated[str, Field(min_length=1)]
            replace: str
            maxN: Annotated[int, Field(default=-1)]

        transformer: Literal["regex.replace"]
        config: "RegexTransformerSchema.Replace.Config"


RegexTransformerSchemas = Union[RegexTransformerSchema.Replace,]



from .config_base import BeetlConfig, ComparisonColumn, SyncConfiguration
from .polar_types import (
    PolarTypeLiterals,
    PolarTypeOverridesParameters,
    get_polar_type_from_literal,
)
from .v1.v1_config import BeetlConfigV1



from typing import Annotated

from pydantic import Field, model_validator

from ...validation import ValidationBaseModel


class SourceConnectionArguments(ValidationBaseModel):
    pass


class SourceConfigArguments(ValidationBaseModel):
    """Class representation of the source connection settings in the beetl config. Used to validate the source configuration settings using pydantic."""

    name: Annotated[str, Field(min_length=1)]
    type: Annotated[str, Field(min_length=1)]
    connection: SourceConnectionArguments

    @model_validator(mode="before")
    def propagate_nested_location(cls, values: dict):
        cls.propagate_location("connection", values)
        return values


class SourceConfig:
    """The connection configuration class used for data sources, abstract"""

    def __init__(cls, arguments: SourceConfigArguments):
        pass
